import os
import inspect
import unittest
import influxdb
from tests.utils.influxdb_test import InfluxDBContainerRequirement, USER, PASSWORD, DB


# import a script with "-" in filename
path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))) + '/../'
with open(path + '/infracheck/checks/influxdb-query', 'r') as script:
    exec(script.read())


class InfluxDBQueryCheckTest(InfluxDBContainerRequirement, unittest.TestCase):
    def test_finds_occurrence_in_json_results(self):
        client = influxdb.InfluxDBClient(host='localhost', port=8086)
        client.switch_database(DB)
        client.switch_user(USER, PASSWORD)

        try:
            for i in range(1, 5):
                client.write_points([
                    {
                        "measurement": "cpu_load_short",
                        "tags": {
                            "host": "server01",
                            "region": "us-west"
                        },
                        "time": "2009-11-10T23:00:{}Z".format(i * 10),
                        "fields": {
                            "value": 0.64 + (i * 10)
                        }
                    }
                ])

            check = InfluxDBQueryCheck(
                host='localhost', port=8086,
                user=USER, password=PASSWORD,
                database=DB
            )

            result, message = check.main('select value from cpu_load_short;', '"value": 10.64')

            self.assertEqual('Query results are matching', message)
            self.assertTrue(result)
        finally:
            client.close()

    def test_will_not_find_occurrence(self):
        client = influxdb.InfluxDBClient(host='localhost', port=8086)
        client.switch_database(DB)
        client.switch_user(USER, PASSWORD)

        try:
            check = InfluxDBQueryCheck(
                host='localhost', port=8086,
                user=USER, password=PASSWORD,
                database=DB
            )

            result, message = check.main('select value from this_metrics_does_not_exist;', '"value": 10.64')

            self.assertEqual('Expected \'[]\' to contain \'"value": 10.64\'', message)
            self.assertFalse(result)
        finally:
            client.close()




# pragma: no cover

import sys
import os
import argparse
import json
from enum import Enum
from .http import HttpServer

t = sys.argv[0].replace(os.path.basename(sys.argv[0]), "") + "/../"

if os.path.isdir(t):
    sys.path.insert(0, t)


from .controller import Controller


class LogLevel(Enum):
    info = 'info'
    debug = 'debug'
    warning = 'warning'
    error = 'error'
    fatal = 'fatal'

    def __str__(self):
        return self.value


def main():
    #
    # Arguments parsing
    #
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--list-available-types',
        help='List all available check types that can be used',
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '--list-all-configurations',
        help='List all configured checks',
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '--list-enabled-configurations',
        help='List only enabled configurations',
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '--directory', '-d',
        help='Alternative project directory',
        default=''
    )
    parser.add_argument(
        '--server-port', '-p',
        help='Server port, default is 7422',
        default=7422
    )
    parser.add_argument(
        '--db-path', '-b',
        help='Database path',
        default='~/.infracheck.sqlite3'
    )
    parser.add_argument(
        '--wait', '-w',
        help='Seconds between doing checks',
        default=0
    )
    parser.add_argument(
        '--timeout', '-t',
        help='Timeout for a single healthcheck to execute',
        default=10
    )
    parser.add_argument(
        '--refresh-time', '-r',
        help='Refresh time in seconds - how often perform health checks. Defaults to 120 (seconds) = 2 minutes',
        default=120
    )
    parser.add_argument(
        '--no-server', '-n',
        help='Do not run the server, just run all the checks from CLI',
        action='store_true'
    )
    parser.add_argument(
        '--server-path-prefix',
        help='Optional path prefix to the routing, eg. /this-is-a-secret will make urls looking like: '
             'http://localhost:8000/this-is-a-secret/',
        default=''
    )
    parser.add_argument(
        '--log-level', '-l',
        help='Logging level name - debug, info, warning, error, fatal. Defaults to "info"',
        type=LogLevel,
        choices=list(LogLevel),
        default='info'
    )
    parser.add_argument(
        '--version', '-v',
        help='Get application version',
        action='store_true'
    )

    parsed = parser.parse_args()
    project_dir = parsed.directory if parsed.directory else os.getcwd()
    server_port = int(parsed.server_port if parsed.server_port else 7422)
    server_path_prefix = parsed.server_path_prefix if parsed.server_path_prefix else ''
    wait_time = int(parsed.wait)
    timeout = int(parsed.timeout)

    app = Controller(
        project_dir=project_dir,
        server_port=server_port,
        server_path_prefix=server_path_prefix,
        db_path=parsed.db_path,
        wait_time=wait_time,
        timeout=timeout,
        log_level=str(parsed.log_level)
    )

    if parsed.version:
        print(app.get_version()['version'])
        sys.exit(0)

    # action: --list-all-configurations
    if parsed.list_all_configurations:
        for name in app.list_enabled_configs():
            print(name)

        sys.exit(0)

    # action: --list-available-types
    if parsed.list_available_types:
        for name in app.list_available_checks():
            print(name)

        sys.exit(0)

    # action: --list-enabled-configurations
    if parsed.list_enabled_configurations:
        for name in app.list_all_configs():
            print(name)

        sys.exit(0)

    # action: perform health checking
    if not parsed.no_server:
        app.spawn_threaded_application(refresh_time=int(parsed.refresh_time))
        server = HttpServer(app=app, port=server_port, server_path_prefix=server_path_prefix)
        server.run()

        sys.exit(0)

    if parsed.no_server:
        result = app.perform_checks().to_hash()
        print(json.dumps(result, sort_keys=True, indent=4, separators=(',', ': ')))

        if not result['global_status']:
            sys.exit(1)

    sys.exit(0)



import sys
import os
import inspect
from datetime import datetime

from rkd.api.testing import BasicTestingCase

path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))) + '/../'
sys.path.insert(0, path)

from infracheck.infracheck.model import ExecutedCheckResult


class ExecutedCheckResultTest(BasicTestingCase):
    def test_from_not_ready(self):
        result = ExecutedCheckResult.from_not_ready('Durruti')

        self.assertEqual(False, result.exit_status)
        self.assertIsNone(result.refresh_time)

    def test_to_hash(self):
        check = ExecutedCheckResult(
            configured_name='Durruti',
            output='Viva la revolution!',
            exit_status=True,
            hooks_output='A las barricadas!'
        )

        check.refresh_time = datetime(2020, 11, 27, 23, 40, 18)  # mock the time
        as_hash: dict = check.to_hash()

        self.assertEqual({
            'checked_at': '2020-11-27 23-40-18',
            'hooks_output': 'A las barricadas!',
            'ident': 'Durruti=True',
            'output': 'Viva la revolution!',
            'status': True
        }, as_hash)



import os
import sys
from typing import List
from .model import ExecutedChecksResultList
from .runner import Runner
from .repository import Repository
from .config import ConfigLoader
from .scheduler import Scheduler
from .versioning import get_version
from rkd.api.inputoutput import IO


class Controller(object):
    """
    Constructs application context and passes actions to given services that are taking care about the processing
    """

    project_dirs: list
    runner: Runner
    repository: Repository
    config_loader: ConfigLoader
    io: IO

    def __init__(self, project_dir: str, server_port: int, server_path_prefix: str,
                 db_path: str, wait_time: int, timeout: int, log_level: str):

        self.io = IO()
        self.io.set_log_level(log_level)
        self.project_dirs = self._combine_project_dirs(project_dir)
        self.config_loader = ConfigLoader(self.project_dirs)
        self.repository = Repository(self.project_dirs, db_path)

        self.runner = Runner(dirs=self.project_dirs, config_loader=self.config_loader,
                             repository=self.repository, timeout=timeout, wait_time=wait_time, io=self.io)

        self.scheduler = Scheduler(self.runner, self.repository, self.io)

    def list_enabled_configs(self) -> List[str]:
        return self.repository.get_configured_checks(with_disabled=False)

    def list_available_checks(self) -> List[str]:
        return self.repository.get_available_checks()

    def list_all_configs(self) -> List[str]:
        return self.repository.get_configured_checks(with_disabled=True)

    def spawn_threaded_application(self, refresh_time: int) -> None:
        """
        Spawns a background worker
        """

        self.scheduler.schedule_jobs_in_background(every_seconds=refresh_time)

    @staticmethod
    def get_version() -> dict:
        """
        Gets Infracheck version
        """

        return {
            "version": get_version(),
            "python": sys.version
        }

    def retrieve_checks(self) -> ExecutedChecksResultList:
        """
        Only retrieves results of last checking
        """

        return self.runner.get_checks_results(self.list_enabled_configs())

    def perform_checks(self) -> ExecutedChecksResultList:
        """
        Runs and returns results synchronously
        """

        configs = self.list_enabled_configs()

        self.runner.run_checks(configs)
        return self.runner.get_checks_results(configs)

    @staticmethod
    def _combine_project_dirs(project_dir: str) -> list:
        paths = [
            # directory specified by eg. the "--directory" commandline parameter
            project_dir,

            # standalone application running from cloned repository
            os.path.dirname(os.path.realpath(__file__)) + '/../',

            # official docker container
            '/app',
            '/data',

            # current directory
            os.getcwd(),
        ]

        return list(filter(lambda path: os.path.isdir(path + '/configured'), paths))




"""
Repository
==========

A data source - database, cache, list of checks, etc.
"""

import os
import posix
import sqlite3
import time
from typing import Union, List
from .model import ExecutedCheckResult
from pickle import loads as deserialize, dumps as serialize
from threading import RLock


class Repository:
    """
    Provides data fetching and caching mechanism
    """

    checks_dirs: list
    configured_dirs: list
    db: sqlite3.Connection
    db_lock: RLock

    def __init__(self, project_dirs: list, db_path: str = '~/.infracheck.sqlite3'):
        self.checks_dirs = []
        self.configured_dirs = []
        self.db_path = db_path
        self.db_lock = RLock()
        self._connect_to_db(db_path)

        for path in project_dirs:
            self.checks_dirs.append(path + '/checks')
            self.configured_dirs.append(path + '/configured')

    def _connect_to_db(self, db_path: str):
        self.db = sqlite3.connect(os.path.expanduser(db_path), check_same_thread=False)

        try:
            self._execute(
                '''
                    CREATE TABLE checks_cache (
                        id INTEGER PRIMARY KEY AUTOINCREMENT, 
                        check_name TEXT, 
                        data TEXT, 
                        date_added TEXT
                    );
                '''
            )
        except sqlite3.OperationalError:
            pass

        try:
            self._execute('CREATE INDEX check_name_index ON checks_cache(check_name);')
        except sqlite3.OperationalError:
            pass

        self.db.commit()

    def get_configured_checks(self, with_disabled: bool = False) -> List[str]:
        found = []

        for path in self.configured_dirs:
            if not os.path.isdir(path):
                continue

            configurations = os.scandir(path)

            for entry in configurations:
                if not with_disabled and not entry.name.endswith('.json'):
                    continue
                elif with_disabled and not (entry.name.endswith('.disabled') or entry.name.endswith('.json')):
                    continue

                found.append(os.path.splitext(entry.name)[0].replace('.json', ''))

        return list(set(found))

    def get_available_checks(self) -> List[str]:
        results = []

        for path in self.checks_dirs:
            available_checks = os.scandir(path)

            def map_checks_to_name(check: posix.DirEntry):
                return str(check.name)

            results += map(map_checks_to_name, available_checks)

        return list(set(results))

    def _purge_cache(self, check_name: str):
        self._execute('DELETE FROM checks_cache WHERE check_name = ?', [check_name])

    def push_to_cache(self, check_name: str, data: ExecutedCheckResult):
        self._purge_cache(check_name)

        self._execute(
            '''
                INSERT INTO checks_cache (id, check_name, data, date_added)
                VALUES (NULL, ?, ?, ?);
            ''',
            [check_name, serialize(data), time.time()]
        )

        self.db.commit()

    def _execute(self, query: str, parameters = None):
        if parameters is None:
            parameters = []

        with self.db_lock:
            return self.db.execute(query, parameters)

    def retrieve_from_cache(self, check_name: str) -> Union[ExecutedCheckResult, None]:
        cursor = self._execute('SELECT data, date_added FROM checks_cache WHERE check_name = ? ' +
                               'ORDER BY date_added DESC LIMIT 1', [check_name])
        data = cursor.fetchone()

        if not data:
            return None

        try:
            return deserialize(data[0])
        except:
            return None




import psycopg2
import psycopg2.extras


class BasePostgreSQL:
    conn = None

    def __init__(self, host: str, dbname: str, port: int, username: str, password: str, connect_timeout: int):
        self.conn = psycopg2.connect(
            database=dbname,
            host=host,
            password=password,
            user=username,
            port=port,
            connect_timeout=connect_timeout
        )

    @staticmethod
    def create_instance(cls):
        return cls(
            host=os.getenv('DB_HOST', 'localhost'),
            dbname=os.getenv('DB_NAME', ''),
            username=os.getenv('DB_USER', 'root'),
            password=os.getenv('DB_PASSWORD', ''),
            connect_timeout=int(os.getenv('DB_CONNECT_TIMEOUT', 10))
        )

    def query(self, qstr: str):
        cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
        cur.execute(qstr)

        return cur.fetchall()

    def validate_replication_row_exists(self, sql: str, expected_status: str, expected_user: str):
        out = self.query(sql)
        active_replications = len(out)
        has_no_access_to_at_least_one_row = False

        for row in out:
            status, conn_info = row

            if conn_info is None:
                has_no_access_to_at_least_one_row = True
                continue

            if status != expected_status and expected_user in conn_info:
                return False, 'Expected "%s" status for conn_info="%s", got "%s"' % (
                    expected_status, conn_info, status
                )

            if status == expected_status and expected_user in conn_info:
                return True, "%i replications active. The replication for '%s' user looks healthy" % (
                    active_replications, expected_user
                )

        if active_replications == 0:
            if has_no_access_to_at_least_one_row:
                return False, "no replications active: possibly connection user has no permissions to view this data"

            return False, "no replications active"

        return False, "%i replications active, but none found for user '%s'" % (active_replications, expected_user)




import os
import json

from .exceptions import ConfigurationException
from .rkd_support import is_rkd_check, rkd_module_exists


class ConfigLoader:
    paths = []

    def __init__(self, paths):
        self.paths = paths

    def load(self, config_name: str) -> dict:
        file_path = self._find_file_path('/configured/', config_name, '.json')

        if not file_path:
            raise FileNotFoundError('Configuration file not found for "' + config_name + '"')

        handle = open(file_path, 'rb')
        content = handle.read()
        handle.close()

        data = json.loads(content)

        self._assert_valid_format(config_name, data)
        self._assert_has_valid_type(str(data['type']))

        return data

    @staticmethod
    def _assert_valid_format(config_name: str, data):
        if "type" not in data:
            raise Exception('Configuration "' + config_name + '" needs to specify a name of a check in field "type"')

    def _assert_has_valid_type(self, type_name: str):
        if is_rkd_check(type_name):
            if not rkd_module_exists(type_name):
                raise ConfigurationException.from_rkd_module_not_existing(type_name)

            return True

        if not self._find_file_path('/checks/', type_name, ''):
            raise ConfigurationException.from_binary_not_found(type_name, self.paths)

        return

    def _find_file_path(self, dir_name: str, config_name: str, suffix: str) -> str:
        for path in self.paths:
            if os.path.isfile(path + dir_name + config_name + suffix):
                return path + dir_name + config_name + suffix

        return ''



import sys
import os
import inspect
from rkd.api.testing import BasicTestingCase

path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))) + '/../'
sys.path.insert(0, path)

from infracheck.infracheck.model import ExecutedChecksResultList, ExecutedCheckResult


class ExecutedChecksResultListTest(BasicTestingCase):
    def test_is_global_status_success(self):
        """
        Checks if global status of the endpoint works as expected in basing on checks results
        """

        with self.subTest('One check is failing, then global status is failure'):
            results = ExecutedChecksResultList()
            results.add('first', ExecutedCheckResult(
                configured_name='first',
                output='Test',
                exit_status=False,
                hooks_output=''
            ))
            results.add('second', ExecutedCheckResult(
                configured_name='second',
                output='Test',
                exit_status=True,
                hooks_output=''
            ))

            self.assertFalse(results.is_global_status_success())

        with self.subTest('All checks are passing, then we have a success'):
            results = ExecutedChecksResultList()
            results.add('first', ExecutedCheckResult(
                configured_name='first',
                output='Test',
                exit_status=True,
                hooks_output=''
            ))
            results.add('second', ExecutedCheckResult(
                configured_name='second',
                output='Test',
                exit_status=True,
                hooks_output=''
            ))

            self.assertTrue(results.is_global_status_success())

        with self.subTest('All checks are failing, then we have a failure'):
            results = ExecutedChecksResultList()
            results.add('first', ExecutedCheckResult(
                configured_name='first',
                output='Test',
                exit_status=False,
                hooks_output=''
            ))
            results.add('second', ExecutedCheckResult(
                configured_name='second',
                output='Test',
                exit_status=False,
                hooks_output=''
            ))

            self.assertFalse(results.is_global_status_success())



import os
import sys
import inspect
from rkd.api.testing import BasicTestingCase


TESTS_PATH = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))) + '/../'
sys.path.insert(0, TESTS_PATH)

from infracheck.infracheck.config import ConfigLoader


class ConfigTest(BasicTestingCase):
    def test_load_finds_file_successfully(self):
        """
        Loads a check configuration with a success
        """

        loader = ConfigLoader([TESTS_PATH + '/example/healthchecks', TESTS_PATH + '/infracheck'])
        check = loader.load('ram')

        self.assertEqual({'type': 'free-ram', 'input': {'max_ram_percentage': '85'}}, check)

    def test_load_does_not_find_file(self):
        loader = ConfigLoader([TESTS_PATH + '/example/healthchecks', TESTS_PATH + '/infracheck'])

        self.assertRaises(FileNotFoundError, lambda: loader.load('not-existing'))

    def test_assert_valid_format(self):
        with self.subTest('Success case'):
            ConfigLoader._assert_valid_format('Durruti', {'type': 'curl'})

        with self.subTest('Failure case - missing "type" attribute'):
            self.assertRaises(Exception, lambda: ConfigLoader._assert_valid_format('Some', {}))
